#!/usr/bin/perl
# $Id: dnsmoncfg.pl,v 1.9.2.1 2008/04/13 15:49:48 ruben Exp $
############################################################################
###    (C)opyright 2003 - 2008 RIPE NCC
###    This file is part of DNSMon
###
###    DNSMon is free software: you can redistribute it and/or modify
###    it under the terms of the GNU General Public License as published by
###    the Free Software Foundation, either version 3 of the License, or
###    (at your option) any later version.
###
###    DNSMon is distributed in the hope that it will be useful,
###    but WITHOUT ANY WARRANTY; without even the implied warranty of
###    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
###    GNU General Public License for more details.
###
###    You should have received a copy of the GNU General Public License
###    along with DNSMon.  If not, see <http://www.gnu.org/licenses/>.
############################################################################


use strict;
use warnings;
use XML::Simple qw(:strict);
use Data::Dumper;
use IO::File;
use Storable;
use Net::DNS::Resolver;
use Carp;
use Socket;
use Socket6;

sub process_config($);
sub create_nsmdconf($$);
sub create_dnsmonconf($$);
sub pserver($$$$$);
sub getservertype($);
sub gettype($);
sub getprofilevalues($$$);
sub try (&@);
sub catch (&);
sub dig($$;$$);


my (
	$verbose,
	$debug,

	@dnsmonattrs_www,
	@dnsmonattrs_nsmd,

	$configname,
	$configpath,
	$webconfigname,
	$probeconfigname,
	$default_domain,
	$dnsmonconfig,
	$probeconfig,

	%tboxes,

	$STEP,
	%server_freqcheck,
	$ERROR,
);

$ERROR =0;

$STEP = 60;		# RRD data interval

$configname		  = 'dnsmoncfg.xml';
$webconfigname    = 'config';
$probeconfigname  = 'nsmd.conf';
$configpath       = '.';
$default_domain	  = 'example.net';
@dnsmonattrs_www  = qw(rrddir pngdir datadir); 
@dnsmonattrs_nsmd = qw(period);

if (system(qw(xmllint --valid --loaddtd --noout ), $configname) == 0)
{
	my $fh = new IO::File($configname,'r');
	my (@result_dnsmonconf,@result_nsmdconf);

	try { 
		$dnsmonconfig = XMLin($fh, ForceArray => 1, KeyAttr => [qw(name type)] );
	} catch {
		die "$_" if $_;
	};

	# print Dumper($dnsmonconfig);

	$probeconfig = process_config($dnsmonconfig);
	# print Dumper($probeconfig);


	@result_dnsmonconf = create_dnsmonconf($dnsmonconfig,$probeconfig);
	@result_nsmdconf = create_nsmdconf($dnsmonconfig,$probeconfig);

	# Check wether the total frequency of measurements is sufficient.
	# Else some plots can't be generated due to lack of data!
	# Inspired by NCC#2007081581

	while (my ($server,$freq) = each %server_freqcheck) {
		if ($freq < $STEP) {
			print "ERROR: server probe total frequency of $server below RRD step ($freq < $STEP). This creates problems for domain plots!\n";
			$ERROR = 1;
		}
	}
	if ($ERROR) {
		print "ERROR: fatal error in preflight phase, correct errors and try again\n";
		exit(1);
	}

	if (@result_dnsmonconf && @result_nsmdconf)
	{
		open (OUT,">${configpath}/${webconfigname}") || die "Can't open ${configpath}/${webconfigname} for writing: $!\n";
		print OUT join("\n",@result_dnsmonconf), "\n";

		open (OUT,">${configpath}/${probeconfigname}") || die "Can't open ${configpath}/${probeconfigname} for writing: $!\n";
		print OUT join("\n",@result_nsmdconf), "\n";

		close(OUT);
	}
}
else
{
	print "ERROR: $configname invalid syntax, unable to create configuration!\n";
	exit(1);
}

sub process_config($)
{
	my ($config) = @_;
	my (%probeconfig);

	# Get the list of domains, always append trailing dot

	foreach my $domain (sort keys %{$config->{domain}})
	{
		my (@servers, %servers, %types);

		# Get configured servers for domain, we look into the NS records for
		# that. Lookup is done through the root and will follow authority rr's
		# down to the right answer.
		# Unless... probing was disabled. only the hardcoded NS set will be used then.
		# however, address records are still being looked up for them

		@servers = map {s/\.$//; $_ } dig($config->{domainalias}{$domain} ? $config->{domainalias}{$domain}{domain} : $domain, "NS", "IN", dig($config->{rootserver},"A")) unless (exists($config->{domain}{$domain}{nsprobe}) && $config->{domain}{$domain}{nsprobe} eq "no");

		# Resolve server aliases, if any
		@{$config->{domain}{$domain}{ns}} = map { $config->{serveralias}{$_} ? $config->{serveralias}{$_}{fqdn} : $_ } @{$config->{domain}{$domain}{ns}};

		# Merge it with the hardwired nameservers for that zone
		push(@servers, map {s/\.$//; $_ } @{$config->{domain}{$domain}{ns}});

		# Lowercase server names and lookup address, currently only V4!
		# First use any hardcoded address, then the regular lookup

		# print Dumper($config->{domain}{$domain}{server}{$_}{addr});

		foreach my $server (@servers) 
		{
			$server =~ tr/A-Z/a-z/;
			push(@{$servers{$server}}, $config->{domain}{$domain}{server}{$server}{addr} || dig($server,"A") || undef ) unless ((defined($config->{domain}{$domain}{nsprobe}) && $config->{domain}{$domain}{nsprobe} eq "v6only") || 
																																	(defined($config->{domain}{$domain}{server}{$server}{proto}) && $config->{domain}{$domain}{server}{$server}{proto} eq "v6only"));
			push(@{$servers{$server}}, $config->{domain}{$domain}{server}{$server}{addr6} || dig($server,"AAAA") || undef ) unless ((defined($config->{domain}{$domain}{nsprobe}) && $config->{domain}{$domain}{nsprobe} eq "v4only") || 
																																	(defined($config->{domain}{$domain}{server}{$server}{proto}) && $config->{domain}{$domain}{server}{$server}{proto} eq "v4only"));

			# strip undefs
			$servers{$server} = [ grep {defined($_)} @{$servers{$server}}];
		}

		while (my ($srv,$addr) = each %servers)
		{
			# Remove unresolvable
			delete $servers{$srv} if ($#{$addr} < 0);
			warn "ERROR: Unresolvable and hardcoded in config for $domain: $srv\n\tEither remove or add an address with <server />" if (($#{$addr} < 0) && grep { /^\Q${srv}\E$/ } @{$config->{domain}{$domain}{ns}} );
		}

		# Probe server characteritics
		# XXX Could be within <server /> but as this is per domain we'll just keep this as it is currently

		foreach my $server (keys %servers)
		{
			foreach my $addr (@{$servers{$server}})
			{
				$types{$addr} = $config->{forcetype}{$server} ? $config->{forcetype}{$server}{type} : getservertype($addr);
			}

			# Expand with profile information, Per server values have precedence over domain values

			# Server has a profile specified ? use it
			my $profile = exists($config->{domain}{$domain}{server}{$server}{profile}) ? $config->{domain}{$domain}{server}{$server}{profile} : $config->{domain}{$domain}{profile};

			# Assign if empty
			foreach my $addr (@{$servers{$server}})
			{
				$probeconfig{$server}{addr}{$addr}{type} ||= $types{$addr};

				# Create a servername configuration mapping, using the server type
				my (%serverconfig) = map { $_ => getprofilevalues($config,$profile,$_) } getquerytype($types{$addr});

				# Store the found configuration per domain per server address
				$probeconfig{$server}{addr}{$addr}{probe}{$domain} = \%serverconfig;
			}

			# Assign if empty, but only when a description is available
			$probeconfig{$server}{description} ||= $config->{serverdescription}{$server}{description} if (exists ($config->{serverdescription}{$server}{description}));
		}
	}
	# Return a reference to the probe configuration for that server
	return \%probeconfig
}

sub create_dnsmonconf($$)
{
	my ($config,$probeconfig) = @_;
	my (%domain_servers, %domain_servers_aliases);
	my (%addr, %addr6);
	my @output;

	my %domain_af = (&AF_INET => "domain", &AF_INET6 => "domain6" );
	my %visibility_af = (&AF_INET => "v4only", &AF_INET6 => "v6only" );

	# Find and use required attributes, complain when they are not present.
	# Don't use defaults, the XML file should provide them

	%domain_servers_aliases = map { $config->{serveralias}{$_}{fqdn} => $_ } keys %{$config->{serveralias}}; 

	foreach my $attr (@dnsmonattrs_www)
	{
		if (exists ($config->{$attr}))
		{
			push(@output,"$attr $config->{$attr}");
		}
		else
		{
			die "$0: required attribute $attr missing from $configname\n";
		}
			
	}

	# Transform probeconfig so we can see which servers belong to which domain.
	# The current order is which domain belong to which server

	while(my ($server, $server_config) = each %{$probeconfig} )
	{
		foreach my $addr (keys %{$server_config->{addr}})
		{
			if (inet_pton(AF_INET,$addr)) {
				$addr{$server} = $addr;
			} elsif (inet_pton(AF_INET6,$addr)) {
				$addr6{$server} = $addr;
			}

			foreach my $domain (keys %{$server_config->{addr}{$addr}{probe}})
			{
				$domain_servers{$domain}{$server} = 1;
			}
		}
	}

	push(@output,"", "#domains", "");

	# Iterate over configured domains
	foreach my $domain (sort keys %domain_servers)
	{
		# XXX Profile specification is mandatory in the XML, so this check is
		# actually superfluous. For safety reasons, I'll just keep it.
		if (exists ($config->{domain}{$domain}{profile}))
		{
			# Store line of text output, if the server has not a profile specified which renders it invisble

			foreach my $af (AF_INET, AF_INET6)
			{
				my @domain_servers;
				my $profilename = $config->{domain}{$domain}{profile};
				my $visibility = getprofilevalues($config,$profilename,"visible");

				# Check visibility
				if ($visibility && ($visibility eq 'no' || ($visibility ne 'yes' && $visibility ne $visibility_af{$af} ) ))
				{
					# This domain will not be visible on the DNSMon website
					# skip it, but mention it when debug mode is on
					warn "$domain: invisble\n" if $debug;
					next; 
				}

				foreach my $domain_server (sort keys %{$domain_servers{$domain}})
				{
					foreach my $addr (sort keys %{$probeconfig->{$domain_server}{addr}})
					{
						next unless inet_pton($af,$addr);

						if (exists($config->{domain}{$domain}{server}{$domain_server}{profile}))
						{
							my $profilename = $config->{domain}{$domain}{server}{$domain_server}{profile};
							$visibility = getprofilevalues($config,$profilename,"visible");
							push (@domain_servers, $domain_server) if not ($visibility && ($visibility eq 'no' || ($visibility ne 'yes' && $visibility ne $visibility_af{$af} ) ));
						}
						else
						{
							# No profile, so probably visible
							push (@domain_servers, $domain_server);
						}
					}
				}

				# Use shortcut in configuration file if present
				@domain_servers = map { exists($domain_servers_aliases{$_}) ? $domain_servers_aliases{$_} : $_ } @domain_servers;

				push(@output, "$domain_af{$af} $domain @domain_servers") if $#domain_servers >= 0; 
			}
		}
	}

	push(@output,"", "# name strings for root servers", "");

	foreach my $name (sort keys %{$config->{serverdescription}})
	{
			push(@output,"sername $name $config->{serverdescription}{$name}{description}") if exists($config->{serverdescription}{$name}{description});
	}

	push(@output,"", "# server aliases", "");
	foreach my $name (sort keys %{$config->{serveralias}})
	{
			push(@output,"seralias $name $config->{serveralias}{$name}{fqdn}");
	}

	# Use ttconfig to have an up to date ttm testbox to country mapping 
	open(TTCONFIG,"ttconfig -v ALL ACTIVE |") || die "Unable to start ttconfig";
	while(defined(my $input = <TTCONFIG>))
	{
		chomp($input);
		my @fields = split(/\s/,$input,3);

		$tboxes{$fields[0]}{$fields[1]}=$fields[2]; 
	}
	close(TTCONFIG);

	push(@output,"", "# country codes for test boxes", "");
	foreach my $name (sort keys %tboxes)
	{
			# XXX default_domain is HARDCODED!!! 
			my $country = exists $tboxes{$name}{country} ? $tboxes{$name}{country} : "??";
			push(@output,"tbname ${name}.${default_domain} $name($country)");
	}

	# Provide convenient shortcuts for the box names
	push(@output,"", "# FQDN 4 test boxes", "");
	foreach my $name (sort keys %tboxes)
	{
			# XXX default_domain is HARDCODED!!! 
			my $fqdn = exists $tboxes{$name}{fqdn} ? $tboxes{$name}{fqdn}: "${name}.${default_domain}";
			push(@output,"tbalias $name ${fqdn}");
	}

	push(@output,"", "# IPv4 addresses 4 nameservers", "");
	push(@output,map { "addr $_ $addr{$_}" } keys %addr);

	push(@output,"", "# IPv6 addresses 4 nameservers", "");
	push(@output,map { "addr6 $_ $addr6{$_}" } keys %addr6);

	return @output;
}

sub create_nsmdconf($$)
{
	my $config = $_[0];
	my (@output, $serverdomains,$serveraddrs, %servers);

	# Find and use required attributes, complain when they are not present.
	# Don't use defaults, the XML file should provide them

	foreach my $attr (@dnsmonattrs_nsmd)
	{
		if (exists ($config->{$attr}))
		{
			push(@output,"$attr $config->{$attr}");
		}
		else
		{
			die "$0: required attribute $attr missing from $configname\n";
		}
	}

	# Iterate over found servers
	foreach my $server (sort keys %$probeconfig)
	{
		# Keep a query type seen list. currently the queries to the chaos realm
		# (CH) are unique per server, whereas SOA queries in the internet (IN)
		# realm are per domain. This is configurable through the profile and
		# thus adjustable per monitored domain


		foreach my $addr (sort keys %{$probeconfig->{$server}{addr}})
		{
			my (%querytypes_seen);

			# Informational comment per server
			push(@output,sprintf("\n# Server %s, \n#     domains %s, type %s\n", $server, join(", ", sort keys %{$probeconfig->{$server}{addr}{$addr}{probe}}),  $probeconfig->{$server}{addr}{$addr}{type}));

			# Iterate over per domain configuration for this server
			foreach my $domain (sort keys %{$probeconfig->{$server}{addr}{$addr}{probe}})
			{
				# Per domain, get the probe parameters
				foreach my $probe ( sort {$config->{typealias}{$a}{type} <=> $config->{typealias}{$b}{type}} keys %{ $probeconfig->{$server}{addr}{$addr}{probe}{$domain} })
				{
					# Create a convenient shortcut, where quite deep in the data structure here
					my $probeparam = $probeconfig->{$server}{addr}{$addr}{probe}{$domain}{$probe};

					# Initialise reference counter to zero
					$querytypes_seen{$probe} = 0 unless exists($querytypes_seen{$probe});

					# skip rest of loop unless we allow multiple items or the
					# reference count is still zero for that query type

					# next unless ( (exists($probeparam->{multiple}) && $probeparam->{multiple} eq "yes") || ! $querytypes_seen{$probe} );

					if ($querytypes_seen{$probe})
					{
						next unless (exists($probeparam->{multiple}) && $probeparam->{multiple} eq "yes");
					}

					# Create the actual configuration line, with on the fly:
					# - Resolving of domain aliases
					# - Decoding the textual query type to a number type that nsmd understands
					push(@output, pserver($server,$addr,$config->{domainalias}{$domain} ? $config->{domainalias}{$domain}{domain} : $domain, $config->{typealias}{$probe}{type},$probeparam));

					# Reference count increase
					$querytypes_seen{$probe}++;
				}
			}
		}
	}
	return(@output);
}

sub pserver($$$$$) {	# print server config lines
	my($server,$addr,$domain,$type,$probeparam) = @_;

	my ($freq,$retry,$timeout) = ($probeparam->{freq}, $probeparam->{retry}, $probeparam->{timeout});

	# For the nsmd program, append a trailing dot if not already present
	$domain =~ s/$/./ unless $domain =~ /\.$/;

	$server_freqcheck{$server} += $freq;

	return sprintf("server %s %s %d %s %d %d %d", $server, $addr, $type, $domain, $freq, $timeout, $retry);
}

sub getquerytype($)
{
	my ($type) = @_;
	my (@queries);

	# Depending on server type, return which query types we are interrested in
	# for that particular server
	# XXX obscure could be better called "unknown" which is more clearer

	if ($type =~ /^bind/ || $type =~ /^obscure/) {  # treat really obscure servers like bind
		@queries = qw(hostname.bind version.bind);
	}
	elsif ($type =~ /^nsd/) {
		@queries = qw(id.server version.server)
	}
	elsif ($type =~ /^all/) {
		@queries = qw(id.server version.bind version.server)
	}

	# Always request for a SOA probe
	push (@queries,"soa");

	return(@queries);
}

sub getservertype($) {	# determine server type
	my($server) = @_;
	my($vers_bind) = '';
	my($vers_server) = '';
	my($id_server) = '';
	my($hostname_bind) = '';


	# check ID queries
	$vers_bind = dig("version.bind.","TXT", "CH", $server);
	$vers_server = dig("version.server.","TXT", "CH", $server);
	$id_server = dig("id.server.","TXT", "CH", $server);
	$hostname_bind = dig("hostname.bind.","TXT", "CH", $server);

	# Little state machine, return server type depending on which lookups in the chaos (CH) realm succeeded

	return("bind (version.bind=$vers_bind hostname.bind=$hostname_bind)") if ($vers_bind && $hostname_bind && !$vers_server && !$id_server);
	return("bind obscure version (hostname.bind=$hostname_bind)") if (!$vers_bind && $hostname_bind && !$vers_server && !$id_server);
	return("bind obscure host (version.bind=$vers_bind)") if ($vers_bind && !$hostname_bind && !$vers_server && !$id_server);
	return("nsd (version.server=$vers_server id.server=$id_server)") if (!$vers_bind && !$hostname_bind && $vers_server && $id_server);
	return("all (hostname.bind=$hostname_bind id.server=$id_server)") if ($hostname_bind && $id_server);
	return("obscure (treated as bind)");
}

sub getprofilevalues($$$)	# Decode profile information, using imports when data cannot be found in the current profile
{
	my ($config,$profile,$setting) = @_;
	my $result = "" ;

	# protect against import loop as the import definition is really a graph
	if (!exists($config->{profile}{$profile}{ROOT}))
	{
		# Protect with sempahore against loop
		$config->{profile}{$profile}{ROOT} = 1;

		# Value we are looking for present in current profile as a setting ? 
		if (exists($config->{profile}{$profile}{settings}{$setting}))
		{
			# Store result of setting
			$result = $config->{profile}{$profile}{settings}{$setting};
		}
		elsif (exists($config->{profile}{$profile}{$setting}))
		{
			# Value we are looking for present in current profile itself ?
			# Store result
			$result = $config->{profile}{$profile}{$setting};
		}
		elsif (exists($config->{profile}{$profile}{import}))
		{
			# if not, and a import field is available, look there
			my $import = $config->{profile}{$profile}{import};

			# recursivly get values
			$result = getprofilevalues($config,$import,$setting);
		}

		# Remove semaphore
		delete $config->{profile}{$profile}{ROOT};
	}
	return $result;
	# return new datastructure
}

# C++ style catch and try, from the perlsub manpage
sub try (&@) {
    my($try,$catch) = @_;
    eval { &$try };
    if ($@) {
        local $_ = $@;
        &$catch;
    }
}
sub catch (&) { $_[0] }

sub dodig(@) {

	my ($name,$type,$domain,$nameserver) = @_;
	my (@answers, $query, $error_string);
    my $res   = Net::DNS::Resolver->new;
	my $retry = 3;

	# Setup packet
	# XXX split(/,\s+/,$type) so we can do IPv4,IPv6  "A,AAAA"
	my $packet = new Net::DNS::Packet( $name, $type, defined($domain) ? $domain : 'IN' );

	# Setup resolver
	$res->debug(1) if $debug;
	$res->nameservers($nameserver) if ($nameserver);
	$res->udp_timeout(5);

	croak("dig: not enough arguments") unless ($name && $type);

	do {
		# As seen from zupd (Actually NCC::Zupd::DNSCache)
		# But disable any die handler that might be present
		try {
			local $SIG{ALRM} = sub { die "timeout" };
			local $SIG{__DIE__} = 'DEFAULT';
			
			# Set watchdog timeout
			alarm( $res->udp_timeout + 1 );

			# clear error_string
			undef $error_string;

			# Send packet
			$query = $res->send( $packet );

			# clear alarm
			alarm 0;
			1;
		} catch {
			/timeout/ && do { 
				$error_string = 'query manually timed out via alarm (implies a library error)'; 
				};
		};

		if ($query) {
			# We had a succesful query. no need to retry, so set to zero
			$retry = 0;

			# Get any set of answers. Either real answers or when that is empty, the data found in the authority section
			# XXX Cache with TTL ? If it is to be done, this is the space and
			# do it like : %answer = map {${_}->rdatastr => time() + ${_}->ttl }
			# Not much refactoring is needed for this.
			@answers = map {${_}->rdatastr } grep { ${_}->type eq $type } (defined($query->answer) && $query->answer ? $query->answer : $query->authority ? $query->authority : () );

			if (!(defined($query->answer) && $query->answer) && $#answers >= 0)
			{
				# We have authorative data. ask a random authorative nameserver for the answer

				my $authorative_ns = $answers[rand($#answers + 1)];
				warn "Querying $authorative_ns for more information about $name,$type,$domain" if $debug;
				# Ok, no answers where found. Let retry with a random nameserver from the authority section
				@answers = dig($name,$type,defined($domain) ? $domain : 'IN', dig($authorative_ns, 'A'));
			}
		}
		else {
			$error_string = "query failed: " . $res->errorstring unless $error_string;
		}

		if ($retry)
		{
			# back off factor while retrying...
			$res->udp_timeout( $res->udp_timeout + 1 );
			warn "$error_string, retrying $retry time(s)for ($name,$type,$domain,$nameserver). timeout set to " . $res->udp_timeout if ($error_string && $debug);
		}
	} while (${retry}--);

	# XXX If this croaks too often, add TTL based cacheing. or fix the network
	carp $error_string if $error_string;
	return @answers;
}

my %cache;

sub dig($$;$$) {

	# Very powerfull short perl for
	# Is the cached value present or assign a reference to it.
	# And oh, return and dereference on the fly please.

	my @result = @{ $cache{"@_"} ||= [dodig(@_)] };

	return wantarray ? @result : $result[0];
}

#	# XXX Uncomment for development purposes only. This will create a disks
#	# persistant cache so you don't have to wait for nameservers and stuff, and can
#	# even develop without any networking
#	
#	BEGIN {
#		use Storable;
#	
#		if (-f '.dnsmon.cache') {
#			%cache = %{ retrieve('.dnsmon.cache') };
#			# print Dumper(\%cache);
#		}
#	
#	}
#	
#	END {
#		use Storable qw(nstore);
#	
#		nstore(\%cache,'.dnsmon.cache');
#	}
