package RIPE::DNSMon::ApacheHandler;
############################################################################
###    (C)opyright 2003 - 2008 RIPE NCC
###    This file is part of DNSMon
###
###    DNSMon is free software: you can redistribute it and/or modify
###    it under the terms of the GNU General Public License as published by
###    the Free Software Foundation, either version 3 of the License, or
###    (at your option) any later version.
###
###    DNSMon is distributed in the hope that it will be useful,
###    but WITHOUT ANY WARRANTY; without even the implied warranty of
###    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
###    GNU General Public License for more details.
###
###    You should have received a copy of the GNU General Public License
###    along with DNSMon.  If not, see <http://www.gnu.org/licenses/>.
############################################################################


use 5.008;

use strict;
use warnings;
no warnings 'uninitialized';

our $VERSION = do { my @r = (q$Revision: 1.11.2.1 $ =~ /\d+/g); sprintf "%d."."%02d" x $#r, @r };

use Apache2::Request;
use Apache2::Const -compile => qw( NOT_FOUND OK REDIRECT DECLINED M_GET);
use Apache2::Cookie;

use HTTP::Date;

use RIPE::Security;
use RIPE::KEY;
use RIPE::KEY::Ticket;

use HTML::Mason::ApacheHandler;

use Cache::FileCache;


use Log::Log4perl 'get_logger';

{
  package HTML::Mason::Commands;
  use Data::Dumper;
  use LockFile::Simple;
  use HTML::Entities;
  use HTML::FillInForm;
  use Log::Log4perl 'get_logger';
  use RIPE::DNSMon::RRD; 
  use RIPE::KEY;
  use RIPE::Security;
  use Time::Piece; 
  use URI;

  our $TICKET;
  our $DEBUG;
  our %CONFIG;

  # Default values. override from httpd.conf
  %CONFIG = (
    conf => 'dnsmon', 
    stem => 'dns-servmon', 
    error_email_to => undef,
    time_delay => 2*(Time::Piece->ONE_HOUR()),    # 2 hour delay
    login_url => 'https://dnsmon-customer.example.net/auth/login', 
    logout_url => 'https://dnsmon-customer.example.net/auth/logout', 
    change_password_url => 'https://dnsmon-customer.example.net/dns-servmon/auth/change_password', 
    public_site_url => 'http://dnsmon.example.net',
    force_cache => 0,
    cache_hold_time => 2*(Time::Piece->ONE_HOUR()),    # cache expires after 2 hours, if enabled.
  );

  sub dnsmon_override_default_config {
      my $r = shift;  
      my $log = get_logger();
      my $conf = $r->dir_config();      # Apache::Table obj, but also a HASH ref

      my @debug_msg;

      for my $k (keys %$conf) {
          next unless $k =~ /^DNSMonConfig(.+)$/;
          my $ck = dnsmon_convert_http_conf_name_to_CONFIG_name($1);
          $CONFIG{$ck} = $conf->get($k);
          push @debug_msg, "$k($ck => $CONFIG{$ck})";
      }
      $log->debug("DNSMonConfig* to \%CONFIG: ".join(", ", @debug_msg));
  }
  sub dnsmon_convert_http_conf_name_to_CONFIG_name {
      my $http_name = shift;
      my $config_name = $http_name;
      $config_name =~ s/(^[A-Z])/lc($1)/e;
      $config_name =~ s/([A-Z])/lc('_'.$1)/ge;
      $config_name;
  }
}

our $ah;
our $cache = new Cache::FileCache( {'namespace' => 'DNSMon_keysession',  'default_expires_in' => 7200, auto_purge_on_set => 1 });

sub handler {
    my $log = get_logger();
    my $r = shift;

    Log::Log4perl::MDC->put("clientip", $ENV{'REMOTE_ADDR'});

    # don't let mason handle these
    return Apache2::Const::DECLINED if $r->uri =~ /(jpg|png|gif|ico|crt|cer|der)$/i;
  
    my %cookies = Apache2::Cookie->fetch;
  
    undef $HTML::Mason::Commands::TICKET;
    my $COOKIE_NAME = RIPE::KEY::CookieName(); 
    if ( $cookies{$COOKIE_NAME} ) {
        my $key = $cookies{$COOKIE_NAME}->value;

	if (!($HTML::Mason::Commands::TICKET = $cache->get($key))) {
		if ($HTML::Mason::Commands::TICKET = RIPE::KEY::Ticket->open($key)) {
			$cache->set($HTML::Mason::Commands::TICKET->id, $HTML::Mason::Commands::TICKET, $HTML::Mason::Commands::TICKET->lifeTime);
		}
	}
        $HTML::Mason::Commands::TICKET
            ? $log->debug("COOKIE $COOKIE_NAME load successful: $key")
            : $log->error("COOKIE $COOKIE_NAME load failure: $key");
    } 
    else {
        $log->debug("no COOKIE $COOKIE_NAME found");
    }

    if ( not defined $ah ) {
        $log->debug("creating Mason ApacheHandler");
        # create it here so we can access the vars defined in httpd.conf via $r
        HTML::Mason::Commands::dnsmon_override_default_config($r);
        $ah = new HTML::Mason::ApacheHandler
        (
            allow_globals=> [qw/$TICKET $DEBUG %CONFIG/],
            decline_dirs=> 1,
            error_mode=> 'fatal',
            error_format=> 'line',
            args_method=> 'mod_perl',
            autohandler_name=> 'autohandler',
            comp_root=> $r->dir_config('DNSMonMasonCompRoot'),
            data_dir=> $r->dir_config('DNSMonMasonDataRoot'),
        );
    }
	# Subscribers are saved from viewing cached data
	if (not $HTML::Mason::Commands::TICKET) {
		$r->no_cache(1);
	} else {
		if ($HTML::Mason::Commands::CONFIG{'customer_only'} && $HTML::Mason::Commands::CONFIG{'public_site_url'} ) {

			# Site in customer mode only. redirect users back from where they came from, but not for the logon/logout urls!

			my $auth_url = new URI($HTML::Mason::Commands::CONFIG{'login_url'});
 
			$log->warn($auth_url->host . " ne " . $r->hostname() . " && " . $auth_url->path . " ne " . $r->uri);
			if ($auth_url->host ne $r->hostname() || $auth_url->path ne $r->uri) {

				$log->warn("Redirecting from ". $r->uri . " to ". $HTML::Mason::Commands::CONFIG{'public_site_url'});

				$r->content_type('text/html; charset=utf-8');
				$r->header_out(Location => $HTML::Mason::Commands::CONFIG{'public_site_url'} );
				return Apache2::Const::REDIRECT;
			}
		} 
		if ($HTML::Mason::Commands::CONFIG{'force_cache'}) {
			# Force cache
			$r->no_cache(0);
			$r->header_out('Expires', HTTP::Date::time2str(time + $HTML::Mason::Commands::CONFIG{'cache_hold_time'}));
			$r->header_out('Cache-Control', "max-age=" . $HTML::Mason::Commands::CONFIG{'cache_hold_time'});
		} else {
			$r->no_cache(1);
		}
	}
	$r->content_type('text/html; charset=utf-8');
  
	return $ah->handle_request($r);
}
1;
