package RIPE::DNSMon::SubscriberManagement;
############################################################################
###    (C)opyright 2003 - 2008 RIPE NCC
###    This file is part of DNSMon
###
###    DNSMon is free software: you can redistribute it and/or modify
###    it under the terms of the GNU General Public License as published by
###    the Free Software Foundation, either version 3 of the License, or
###    (at your option) any later version.
###
###    DNSMon is distributed in the hope that it will be useful,
###    but WITHOUT ANY WARRANTY; without even the implied warranty of
###    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
###    GNU General Public License for more details.
###
###    You should have received a copy of the GNU General Public License
###    along with DNSMon.  If not, see <http://www.gnu.org/licenses/>.
############################################################################

use strict;
use warnings;
use Error;

use RIPE::Security;
use RIPE::KEY::DataStoreDBI;
use Log::Log4perl qw(get_logger);

=head1 NAME

RIPE::DNSMon::SubscriberManagement - Manage DNSMon subscribers

=head1 SYNOPSIS

Local usage:

use RIPE::DNSMon::SubscriberManagement

@groups = RIPE::DNSMon::SubscriberManagement->groups($user);
$pass = RIPE::DNSMon::SubscriberManagement->user($user);

SOAP Usage:

server side:

<Location /soap>
        AllowOverride None
        Order deny,allow
        Allow from 193.0.0.0/21
        Deny from all

        SetHandler perl-script
        PerlHandler Apache::SOAP
        PerlSetVar dispatch_to "RIPE::DNSMon::SubscriberManagement"
        PerlSetVar options "compress_threshold => 10000"
</Location>

client side

my $soap = SOAP::Lite
    ->uri('http://server.example.net/RIPE/DNSMon/SubscriberManagement')
    ->proxy('http://server.example.net/soap/', keep_alive => 1);

@groups = $soap->groups($user)->paramall;
$pass = $soap->user($user)->result;

=head1 DESCRIPTION

The RIPE::DNSMon::SubscriberManagement module manages the accounts in DNSMon as
the RIPE::Security framework does not foresee in this by design.

It uses RIPE::Security to get a handle to the backend.

=head1 METHODS

=over

=cut

BEGIN {
    use Exporter   ();
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);

    # set the version for version checking
    $VERSION = do { my @r = (q$Revision: 1.2.4.1 $ =~ /\d+/g); sprintf "%d."."%02d" x $#r, @r }; # must be all one line, for MakeMaker

    @ISA         = qw(Exporter);
    @EXPORT      = ( );
    %EXPORT_TAGS = ( );     # eg: TAG => [ qw!name1 name2! ],

    # your exported package globals go here,
    # as well as any optionally exported functions
    @EXPORT_OK   = ( );
}
our @EXPORT_OK;
our $ctx;

# XXX FIXME allow override
my $login_modules_conf = RIPE::Security::Config->loginModules;
$login_modules_conf->{dnsmon} =
[
    {
        class => 'RIPE::Security::LoginModule::DNSMon',
        type => 'auth',     # auth, account, session, password
        flag => 'required', # required, requisite, sufficient, optional
        options =>
        {   
            dbi => 
            [
                'dbi:SQLite:dbname=/home/ttraffic/DNSMon_Users', 
                '', 
                ''
            ],
        },
    }
];

&initialize();

=item C<initialize()>

Initializes L<RIPE::Security::LoginContext> on package loading time using dummy
callback and subject handles and calling the private method _initialize_login_modules

you do not need to call this function yourself.

=cut


sub initialize {
	my ($self) = shift;
	my $callback = new RIPE::Security::CallbackHandler::Passive;
	my $subject = new RIPE::Security::Subject;
	$ctx = new RIPE::Security::LoginContext("dnsmon", $subject, $callback);
	$ctx->_initialize_login_modules;
}

=item C<_dbh()>

Internal access to RIPE::Security::LoginContext's database backend

Returns: L<DBI> handle

=cut

sub _dbh {
	$ctx->{lm_conf}->[0]->{lm}->{rdbm}->_dbh();
}

=item C<users()>

Return a list of all users in the system as an array

=cut

sub users {
	my ($self) = shift;
	my @users = map { @{$_}} @{_dbh()->selectall_arrayref("select logname from Users")};
	return @users;
}

=item C<user(I<$user>)>

Test for existance of I<$user>. 

I<$user> - Test this username for existance

Returns: password when I<$user> exists, undef otherwise

Exception: throws I<Error::Simple> on input validation errors

=cut

sub user {
	my ($self) = shift;
	my ($user) = $_[0] =~ /^([\w\-_\.]+)$/;
	my ($password);
	if ($user) {
		$password = _dbh()->selectrow_array("select password from Users where logname = ?", undef, $user);
	} else {
		throw Error::Simple("username required");
	}
	
	return $password;
}

=item C<groups(I<$user>)>

Get groups I<$user> is member of.

I<$user> - Get all configured groups for this member

Returns: all groups as an array or the empty list

Exception: throws I<Error::Simple> on input validation errors

=cut

sub groups($) {
	my ($self) = shift;
	my ($user) = $_[0] =~ /^([\w\-_\.]+)$/;
	my @groups = ();
	if ($user) {
		@groups = map { @{$_}} @{_dbh()->selectall_arrayref("select groups from Groups where logname = ?", undef, $user)};
	} else {
		throw Error::Simple("username required");
	}
	return @groups;
}

=item C<adduser(I<$user>,I<$password>)>

Add or update an account

I<$user> - user to be added or if it exists, modified.

I<$password> - password to be set or changed.

Returns: boolean indicating success or failure

Exception: throws Error::Simple for either input validation errors or database errors

=cut

sub adduser($$) {
	my ($self) = shift;
	my ($user,$password) = map {/^([\w\-_\.]+)$/} @_;
	my ($result) = 0;
	if ($user && $password) {
		my $query;
		if ($self->user($user)) {
			$query = "update Users set password = ? where logname = ? ";
		} else {
			$query = "insert into Users(password,logname) values(?,?)";
		}
		if ($result = _dbh()->do($query, undef, $password, $user)) {
			_dbh->commit() unless _dbh->{AutoCommit};
		} else {
			throw Error::Simple("\"$query\" failed for (XXXX,$user) database error: \"" . _dbh()->errstr . "\"");
		}
	} else {
		throw Error::Simple("username required");
	}
	return $result;
}

=item C<addusers(I<%users>)>

Add or update an account in bulk, specifically for SOAP

I<%users> - hash containing usernames for the keys and passwords for the values

Returns: boolean indicating success or failure

Exception: throws Error::Simple for either input validation errors or database errors (propagated from adduser)

=cut

sub addusers(%) {
	my ($self) = shift;
	my $result = 0;
	my %accounts = @_;

	while (my ($user,$password) = each %accounts) {
		if (adduser($user,$password)) {
			$result++;
		} else {
			throw Error::Simple("adduser($user,XXXX) failed");
		}
	}
	return $result;
}

=item addgroups(I<$user>,I<@groups>)

Add groups to an existant user

I<$user> - User to be added to the specified groups

I<@groups> - list of groups this user should be a member of. Existing membership will be silently ignored

Returns: boolean indicating success or failure

Exception: throws Error::Simple for either input validation errors or database errors

=cut

sub addgroups($@) {
	my ($self) = shift;
	my ($user,@groups) = map {/^([\w\-_\.]+)$/} @_;
	my $result = 0;
	if ($self->user($user)) {
		my (@newgroups) = grep { !/^(@{[ join("|",$self->groups($user)) ]})$/ } @groups;
		if (@newgroups) {
			my $query = "insert into Groups(logname,groups) values(?,?)";
			my $stm = _dbh->prepare($query);
			if ($stm) {
				foreach my $newgroup (@newgroups) {
					if ($stm->execute($user,$newgroup)) {
						$result++;
					} else {
						# Query failed, rollback if autocommit was off
						# XXX you're screwed with autocommit on
						_dbh->rollback() unless _dbh->{AutoCommit};
						throw Error::Simple("query \"$query\" failed for ($user,$newgroup) database error: \"" . _dbh()->errstr . "\"");
					}
				}
				$stm->finish;
				if ($result < ($#newgroups + 1)) {
					_dbh->rollback() unless _dbh->{AutoCommit};
					throw Error::Simple("query \"$query\" failed, number of updated rows not equal to groups present. database error: \"" . _dbh()->errstr . "\"");
				} else {
					_dbh->commit() unless _dbh->{AutoCommit};
				}
			}
		}
	} else {
		throw Error::Simple("addgroups: nonexistant user $user");
	}
	
	return $result;
}

=item C<delgroups(I<$user>,I<@groups>)>

Removes groups from an existant user

I<$user> - User to be removed from the specified groups

I<@groups> - list of groups this user should be stripped from. non membership groups will be silently skipped

Returns: boolean indicating success or failure

Exception: throws Error::Simple for either input validation errors or database errors


=cut

sub delgroups($@) {
	my ($self) = shift;
	my ($user,@groups) = map {/^([\w\-_\.]+)$/} @_;
	my $result = 0;
	if ($self->user($user)) {
		my $query = sprintf("delete from Groups where logname = %s and groups in (%s)",_dbh->quote($user), join(",", map { _dbh->quote($_) } @groups));
		$result = _dbh->do($query);
		throw Error::Simple("query \"$query\" failed: \"" . _dbh()->errstr . "\"");
	} else {
		throw Error::Simple("nonexistant user $user");
	}
	return $result;
}

=item C<delusers(I<@users>)>

Deletes users from the sytem, after revoking membership of all configured groups

I<@users> - Array of usernames to be deleted

Returns: boolean indicating success or failure

Exception: throws Error::Simple for either input validation errors or database errors

=cut

sub delusers(@) {
	my ($self) = shift;
	my (@users) = map {/^([\w\-_\.]+)$/} @_;
	my (@groups);
	my $result = 0;

	foreach my $user (@users) {
		if (@groups = $self->groups($user)) {
			last unless $self->delgroups($user,@groups);
		}
		if (_dbh->do("delete from Users where logname = ?", undef, $user)) {
			$result++;
		} else {
			throw Error::Simple("query \"delete from Users where logname = ?\" for user $user failed, database error: \"" . _dbh()->errstr . "\"");
		}
	}
	if ($result) {
		_dbh->commit() unless _dbh->{AutoCommit};
	}
	return $result;
}

1;
__END__

=back

=head1 NOTES

All possible parameters are untainted with the following regular expression: C</^([\w\-_\.]+)$/>

Because of the possible usage through SOAP::Lite, the module calls are to be
method style (FOO->bar() ) instead of function style (FOO::bar() ).

calling these methods as functions will have an undefined result.

RIPE::DNSMon::SubscriberManagement does tap into RIPE::Security to get
access to the SQL database which is used as a backing store for the account
information. This interface is not documented.

This module uses L<Error> to throw exceptions

=cut
