/*
 *   (C)opyright 2003 - 2008 RIPE NCC
 *   This file is part of DNSMon
 *
 *   DNSMon is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   DNSMon is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with DNSMon.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>

#include "nsmd.h"
#include "dns.h"

/* #define DNS_DEBUG */

#ifndef NS_CMPRSFLGS /* only on newer FreeBSD's */
#define NS_CMPRSFLGS     0xc0    /* Flag bits indicating name compression. */
#endif /* NS_CMPRSFLGS */

#define DNS_HDR_SIZE 12

#define NS_MAXCDNAME     255     /* maximum compressed domain name */

int dns_id;

typedef struct dns_header {
    u_int16_t ID;         /* identifier */        
    u_int16_t QR:1,       /* query / response */
              OPCODE:4,   /* query type */
              AA:1,       /* Authoritive Answer */
              TC:1,       /* TrunCation */
              RD:1,       /* Recursion Desired */
              RA:1,       /* Recursion Available */
              Z:3,        /* Reserved */
              RCODE:4;    /* Response Code */
    u_int16_t QDCOUNT;    /* number of question entries */
    u_int16_t ANCOUNT;    /* number of answers */
    u_int16_t NSCOUNT;    /* number of resource records */
    u_int16_t ARCOUNT;    /* number of additional records */
} __attribute__((packed)) dns_header_t;

struct dns_query {
//    char * NAME;
    u_int16_t TYPE;
    u_int16_t CLASS;
} __attribute__((packed));

struct dns_answer {
//    char * NAME;
    u_int16_t TYPE;
    u_int16_t CLASS;
    u_int32_t TTL;
    u_int16_t RD_LENGTH;
//    char * RDATA;
} __attribute__((packed));

/* query types, internal value */
#define QT_HOSTNAME_BIND 1
#define QT_ID_SERVER 2
#define QT_VERSION_BIND 3
#define QT_VERSION_SERVER 4
#define QT_SOA 5

int parse_dns_header_ntohs(dns_header_t *);
int extractquestion(char *,int);
int extractanswer(char *,int n, char * result);
int extractauthoritive(char *,int);
int extractadditional(char *,int);
int extractdomainname(char *, char *, int);
int makequery(char *query,char *lookupname, u_int16_t qtype, u_int16_t qclass);
u_int32_t get32b (char *p);
void my_name_ntop(const u_char *src, char *dst, size_t srcsiz, size_t dstsiz);
int ns_name_unpack(const u_char *msg, const u_char *eom, const u_char *src,
            u_char *dst, size_t dstsiz);

int
buildpacket(char * packet, struct server list[], int entry)
{
    dns_header_t dh;
    char *query = (char *)malloc(260);
    int querysize = 0;
    char lookupname[19];

    bzero(packet, 512);

    switch (list[entry].query_type) {
    case QT_HOSTNAME_BIND:
        strcpy(lookupname, "hostname.bind.\0");
        strcpy(list[entry].qry, "hostname.bind");
        querysize = makequery(query, lookupname, T_TXT, C_CHAOS);
        break;
    case QT_ID_SERVER:
        strcpy(lookupname, "id.server.\0");
        strcpy(list[entry].qry, "id.server");
        querysize = makequery(query, lookupname, T_TXT, C_CHAOS);
        break;
    case QT_VERSION_BIND:
        strcpy(lookupname, "version.bind.\0");
        strcpy(list[entry].qry, "version.bind");
        querysize = makequery(query, lookupname, T_TXT, C_CHAOS);
        break;
    case QT_VERSION_SERVER:
        strcpy(lookupname, "version.server.\0");
        strcpy(list[entry].qry, "version.server");
        querysize = makequery(query, lookupname, T_TXT, C_CHAOS);
        break;
    case QT_SOA:
        /* if domain == ".", make the query empty " */
        if (!strncmp(list[entry].domain, ".", 1) &&
                strlen(list[entry].domain) == 1)
            strcpy(lookupname, "");
        else
            strcpy(lookupname, list[entry].domain);
        strcpy(list[entry].qry, "soa");
        querysize = makequery(query, lookupname, T_SOA, C_IN);
        break;
    default:
        printf("ERROR: %s() unknown query type\n", __FUNCTION__);
        exit(1);
        /* NOT REACHED */
    }

    /* 
     * query identifier is 16 bit, if it outgrows, wrap it
     */
    if(++dns_id > 0xffff)
        dns_id = 0;
    dh.ID = htons(0xffff & dns_id);

    /* query flags */
    dh.QR = 0;
    dh.OPCODE = 0; 
    dh.AA = 0;
    dh.TC = 0; 
    dh.RD = 0;   
    dh.RA = 0;   
    dh.Z = 0;    
    dh.RCODE = 0;  

    dh.QDCOUNT = htons(1);
    dh.ANCOUNT = htons(0);
    dh.NSCOUNT = htons(0);
    dh.ARCOUNT = htons(0);

    /* copy header to beginning of packet */
    memcpy(packet, &dh, DNS_HDR_SIZE);
    list[entry].size = DNS_HDR_SIZE;

    /* append query a after header */
    memcpy(packet+DNS_HDR_SIZE, query, querysize);

    free(query);

    list[entry].size += querysize;

    return (dns_id);
}

int
makequery(char *query, char *lookupname, u_int16_t qtype, u_int16_t qclass)
{
    int i = 0;
    char *myname = (char *)malloc(100);
    char *myname_tmp;
    char *token;
    struct dns_query *rr;

    /* store this pointer as it will be changed by strsep later on */
    myname_tmp = myname;

    strcpy(myname, lookupname);

    while (strlen(myname) > 0) {
        token = strsep(&myname, ".");
#ifdef DNS_DEBUG
        printf("token: \"%s\"\n", token);
        printf("strlen(token)=%d\n", strlen(token));
#endif /* DNS_DEBUG */
        *(query+i++) = strlen(token);
        memcpy(query+i, token, strlen(token));
        i += strlen(token);
#ifdef DNS_DEBUG
        printf("myname: \"%s\"\n", myname);
        printf("strlen(myname)=%d\n", strlen(myname));
#endif /* DNS_DEBUG */
    }
    *(query+i++) = '\0';

    /* set query type and class */
    rr = (struct dns_query *)(query+i);
    rr->TYPE = htons(qtype);
    rr->CLASS = htons(qclass);

    i += 4; /* TYPE + CLASS */

    free(myname_tmp);

    return (i);
}
int
parse_udp_packet(char *packet, char * result)
{
    dns_header_t *dh;
    int n;
    int i = 0;
    int id;

    dh = (dns_header_t *) packet;

#ifdef DNS_DEBUG
    printf("Parsing header:\n");
#endif /* DNS_DEBUG */
    id = parse_dns_header_ntohs(dh);

    i = DNS_HDR_SIZE; // skip header

#ifdef DNS_DEBUG
    if (dh->QDCOUNT)
        printf("Parsing question(s):\n");
#endif /* DNS_DEBUG */
    for (n = 0; n < ntohs(dh->QDCOUNT); n++)
        i = extractquestion(packet, i);

#ifdef DNS_DEBUG
    if (dh->ANCOUNT)
        printf("Parsing answer(s):\n");
#endif /* DNS_DEBUG */
    for (n = 0; n < ntohs(dh->ANCOUNT); n++)
        i = extractanswer(packet, i, result);
    if (!dh->ANCOUNT)
        strcpy(result, "UNKNOWN");

#ifdef DNS_DEBUG
    if (dh->NSCOUNT)
        printf("Parsing authority record(s):\n");
#endif /* DNS_DEBUG */
    for (n = 0; n < ntohs(dh->NSCOUNT); n++)
        i = extractauthoritive(packet, i);

#ifdef DNS_DEBUG
    if (dh->ARCOUNT)
        printf("Parsing additional record(s):\n");
#endif /* DNS_DEBUG */
    for (n = 0; n < ntohs(dh->ARCOUNT); n++)
        i = extractadditional(packet, i);
    
    return (id);
}

int
parse_raw_packet(char *buf, char * result)
{
    /* Don't do anything here now, save the best for later ;-) */

    return (0);
}

int
extractquestion(char *query_p, int n)
{
    char *name = (char *)malloc(NS_MAXCDNAME);
    int i = 0;
    struct dns_query *rr;

    i = extractdomainname(query_p, name, n+i);
#ifdef DNS_DEBUG
    printf("rr->NAME: %s\n", name);
#endif /* DNS_DEBUG */
    free(name);

    rr = (struct dns_query *)(query_p+i+n);
#ifdef DNS_DEBUG
    printf("rr->TYPE: 0x%.4x\n", ntohs(rr->TYPE));
    printf("rr->CLASS: 0x%.4x\n", ntohs(rr->CLASS));
#endif /* DNS_DEBUG */
    i += 4; /* TYPE + CLASS */

    return (i+n);
}

int
extractanswer(char *query_p, int n, char * result)
{
    int i = 0;
    char *qname = (char *)malloc(NS_MAXCDNAME); /* freed later */
    char *aname = (char *)malloc(NS_MAXCDNAME); /* freed later */
    struct dns_answer *rr;

    bzero(qname, NS_MAXCDNAME);
    bzero(aname, NS_MAXCDNAME);

    i = extractdomainname(query_p, qname, n+i);
#ifdef DNS_DEBUG
    printf("rr->QNAME: %s\n", qname);
#endif /* DNS_DEBUG */
    free(qname);

    rr = (struct dns_answer *)(query_p+i+n);

#ifdef DNS_DEBUG
    printf("rr->TYPE: 0x%.4x\n", ntohs(rr->TYPE));
    printf("rr->CLASS: 0x%.4x\n", ntohs(rr->CLASS));
    printf("rr->TTL: 0x%.8lx\n", ntohl(rr->TTL));
    printf("rr->RD_LENGTH: 0x%.4x\n", ntohs(rr->RD_LENGTH));
#endif /* DNS_DEBUG */

    i += 10; /* TYPE + CLASS + TTL + RD_LENGTH */

#ifdef DNS_DEBUG
	printf("extractanswer(): i: %d\n", i);
#endif /* DNS_DEBUG */

    switch (ntohs(rr->TYPE)) {
        u_int32_t serial;
    case T_TXT: /* TEXT */

        /* TXT is actually not a label!! */
        strncpy(result, (query_p+n+i+1), *(query_p+n+i));

#ifdef DNS_DEBUG
        printf("XXXX: TXT: %s\n", result);
#endif /* DNS_DEBUG */

        i += ntohs(rr->RD_LENGTH);

        break;
    case T_SOA: /* SOA */
        i += extractdomainname(query_p, aname, n+i);
        i += extractdomainname(query_p, aname, n+i);

        serial = get32b(query_p+n+i);
        i += 4; /* serial */
        sprintf(result, "%d", serial);

        break;
    default:
        printf("ERROR: %s(): unknown type of packet received: %d\n", 
            __FUNCTION__, ntohs(rr->TYPE));
        break;
    }

    free(aname);

    return (i+n);
}

int
extractauthoritive(char *query_p, int n)
{
    char *name = (char *)malloc(NS_MAXCDNAME);
    int i = 0;
    struct dns_query *rr;

    i = extractdomainname(query_p, name, n+i);
#ifdef DNS_DEBUG
    printf("rr->NAME: %s\n", name);
#endif /* DNS_DEBUG */
    free(name);

    rr = (struct dns_query *)(query_p+i+n);
#ifdef DNS_DEBUG
    printf("rr->TYPE: 0x%.4x\n", ntohs(rr->TYPE));
    printf("rr->CLASS: 0x%.4x\n", ntohs(rr->CLASS));
#endif /* DNS_DEBUG */
    i += 4; /* TYPE + CLASS */

#ifdef DNS_DEBUG
    printf("i at end of extractauthoritive: %d\n", i);
#endif /* DNS_DEBUG */

    return (i+n);
}

int
extractadditional(char *query_p,int i)
{
                
#ifdef DNS_DEBUG
    printf("i at end of extractadditional: %d\n", i);
#endif /* DNS_DEBUG */

    return (i);
}

int
extractdomainname(char * packet, char *domainname, int p)
{
    int len;
    u_char tmp[NS_MAXCDNAME];

    bzero(domainname, NS_MAXCDNAME);
    bzero(&tmp, NS_MAXCDNAME);

    len = ns_name_unpack(packet, packet + NS_MAXCDNAME, packet + p, tmp,
                        sizeof tmp);

    my_name_ntop(tmp, domainname, len, NS_MAXCDNAME);

    return (len);
}

/*
 *  DNS routine
 *  Convert encoded domainname to printable ascii name
 */
void
my_name_ntop(const u_char *src, char *dst, size_t srcsiz, size_t dstsiz)
{
    int len = 0;
    int i = 0;

    bzero(dst, dstsiz);

    while (*(src+i) != 0) {
        /* length byte and increase 'i' to first char */        
        len = *(src+i++);
        /* copy len bytes */
        strncat(dst, src+i, len);
         /* increase i with len to jump to next label */
        i += len;
         /* if not end of domainname, put a dot between labels */
        if (*(src+i) != 0)
            strcat(dst, "."); 
    }
}

/*
 *  Debugging routine to display DNS header
 */
int
parse_dns_header_ntohs(dns_header_t *dh)
{
    int id;

    id = ntohs(dh->ID);
#ifdef DNS_DEBUG
    printf("ID:%d\n", id);
    printf("FLAGS:0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x \n", 
        dh->QR, dh->OPCODE, dh->AA, dh->TC, 
        dh->RD, dh->RA, dh->Z, dh->RCODE);
    printf("QDCOUNT:%d\n", ntohs(dh->QDCOUNT));
    printf("ANCOUNT:%d\n", ntohs(dh->ANCOUNT));
    printf("NSCOUNT:%d\n", ntohs(dh->NSCOUNT));
    printf("ARCOUNT:%d\n", ntohs(dh->ARCOUNT));
#endif /* DNS_DEBUG */

    return (id);
}
 
/* get 4 bytes from memory
 * eg.  used to extract serial number from soa packet
 */
u_int32_t
get32b (char *p)
{
    u_int32_t var;

    var = (0x000000ff & *(p)) << 24;
    var |= (0x000000ff & *(p+1)) << 16;
    var |= (0x000000ff & *(p+2)) << 8;
    var |= (0x000000ff & *(p+3));
        
    return (var);
}

/*
 * ns_name_unpack(msg, eom, src, dst, dstsiz)
 *    Unpack a domain name from a message, source may be compressed.
 * return:
 *    -1 if it fails, or consumed octets if it succeeds.
 */
int
ns_name_unpack(const u_char *msg, const u_char *eom, const u_char *src,
           u_char *dst, size_t dstsiz)
{
    const u_char *srcp, *dstlim;
    u_char *dstp;
    int n, len, checked;

    len = -1;
    checked = 0;
    dstp = dst;
    srcp = src;
    dstlim = dst + dstsiz;
    if (srcp < msg || srcp >= eom) {
        errno = EMSGSIZE;
        return (-1);
    }
    /* Fetch next label in domain name. */
    while ( ((n = *srcp++) != 0) ) {
        /* Check for indirection. */
        switch (n & NS_CMPRSFLGS) {
        case 0:
            /* Limit checks. */
            if (dstp + n + 1 >= dstlim || srcp + n >= eom) {
                errno = EMSGSIZE;
                return (-1);
            }
            checked += n + 1;
            *dstp++ = n;
            memcpy(dstp, srcp, n);
            dstp += n;
            srcp += n;
            break;

        case NS_CMPRSFLGS:
            if (srcp >= eom) {
                errno = EMSGSIZE;
                return (-1);
            }
            if (len < 0)
                len = srcp - src + 1;
            srcp = msg + (((n & 0x3f) << 8) | (*srcp & 0xff));
            if (srcp < msg || srcp >= eom) {  /* Out of range. */
                errno = EMSGSIZE;
                return (-1);
            }
            checked += 2;
            /*
             * Check for loops in the compressed name;
             * if we've looked at the whole message,
             * there must be a loop.
             */
            if (checked >= eom - msg) {
                errno = EMSGSIZE;
                return (-1);
            }
            break;

        default:
            errno = EMSGSIZE;
            return (-1);            /* flag error */
        }
    }
    *dstp = '\0';
    if (len < 0)
        len = srcp - src;
    return (len);
}
