/*
 *   (C)opyright 2003 - 2008 RIPE NCC
 *   This file is part of DNSMon
 *
 *   DNSMon is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   DNSMon is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with DNSMon.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <sys/wait.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <pwd.h>
#include <signal.h>

#ifdef __FreeBSD___
#include <libutil.h>
#endif /* __FreeBSD__ */
#ifdef __linux__
#include "setproctitle.h"
#endif /* __linux__ */



#include "nsmd.h"
#include "get_ip_address.h"
#include "receiving_process.h"
#include "sending_process.h"
#include "process_data.h"
#include "socket.h"

void setup_signal_handlers();
void handle_signals(int signo);

#ifdef IPV6_SUPPORT
int candov6 = 0;
#endif
int mydebug = 0;

struct globaltime gt;
pid_t recv_pid, send_pid, proc_pid;

int
main(int argc, char *argv[])
{
    struct passwd * pwinfo;
    int wait_status;
    pid_t wait_pid;

#ifdef IPV6_SUPPORT
	struct etherifdata eid;
	char errbuf[256];
#endif

    if (argc > 1) {
        if (!strncmp(argv[1], "-d", 2))
            mydebug = 1;
	else if (!strncmp(argv[1], "-v", 2)) {
#ifdef IPV6_SUPPORT
            printf("nsmd IPv4 and IPv6 version\n");
#else
            printf("nsmd IPv4 version\n");
#endif
            exit(0);
        } else {
            printf("ERROR: unknown option\n");
            exit (1);
        }
    }

#ifdef IPV6_SUPPORT
	while (!candov6 && (NextEthernetIFInfo(&eid,errbuf,ETHIFINFO_ALIASES_TOO) == 0))
		/* Does the box have a configured IPv6 address ? then we can send IPv6 traffic out */
		candov6 = !IN6_IS_ADDR_UNSPECIFIED(&eid.ethif_addr6);
	EndEthernetIFInfo();
#endif

    if (geteuid() > 0) {
        printf("ERROR: effective UID needs to be 0\n");
        exit(1);
    }

    /* we need this for config and output dir */
    if (!getenv("TTROOT")) {
        printf("ERROR: TTROOT environment variable not set!\n");
        exit(1);
    }

    /* XXX - daemonize */
    if (!mydebug)
        daemon(0, 0);

    /* set argv[0] to identify process more easily */
    setproctitle("[controlling process]");

    /* get the first real ip address we encounter */
    get_ip_address();

    /* initialize connect structures for all targets */
    setup_udp_socket(); 
    setup_raw_socket(); 
#ifdef IPV6_SUPPORT
    setup_udp_v6_socket();
#endif /* IPV6_SUPPORT */

    /* drop privileges */
    pwinfo = getpwnam("ttraffic");
    if (pwinfo) {
        setuid(pwinfo->pw_uid);
    } else {
        printf("ERROR: failed to drop privileges\n");
        exit(1);
    }

    /* setup signal handling */
    setup_signal_handlers();

    /* fork of receiver process */
    recv_pid = receiving_process();

    /* fork of sender process */
    send_pid = sending_process();

    /* fork of data processing process */
    proc_pid = process_data();

    printf("pids: send=%d recv=%d proc=%d control=%d\n", send_pid,
        recv_pid, proc_pid, getpid());

    /* if any of our children die, die too */
    wait_pid = wait(&wait_status);
    printf("After wait: %d\n", wait_pid);

    if(wait_pid == send_pid) {
        printf("Sending process died\n");
        kill(proc_pid, SIGTERM);
        kill(recv_pid, SIGTERM);
    } else if(wait_pid == proc_pid) {
        printf("Data processing died\n");
        kill(send_pid, SIGTERM);
        kill(recv_pid, SIGTERM);
    } else if(wait_pid == recv_pid) {
        printf("Receiving process died\n");
        kill(proc_pid, SIGTERM);
        kill(send_pid, SIGTERM);
    } else {
        printf("Unknown wait result\n");
    }

    /* close socket foo */
    breakdown_sockets();

    return (0);
}

void
setup_signal_handlers()
{
    signal(SIGHUP, handle_signals);
    signal(SIGINT, handle_signals);
    signal(SIGQUIT,handle_signals);
    signal(SIGILL, handle_signals);
    signal(SIGTRAP,handle_signals);
    signal(SIGABRT,handle_signals);
#ifndef __linux__
    signal(SIGEMT, handle_signals);
#endif /* __linux__ */
    signal(SIGFPE, handle_signals);
    signal(SIGBUS, handle_signals);
    signal(SIGSEGV,handle_signals);
    signal(SIGSYS, handle_signals);
    signal(SIGPIPE,handle_signals);
    signal(SIGALRM,handle_signals);
    signal(SIGTERM,handle_signals);
    signal(SIGXCPU,handle_signals);
    signal(SIGXFSZ,handle_signals);
    signal(SIGVTALRM,handle_signals);
    signal(SIGPROF,handle_signals);
    signal(SIGUSR1,handle_signals);
    signal(SIGUSR2,handle_signals);
}

void
handle_signals(int signo)
{
    printf("Received signal: %d\n", signo);

    printf("Killing children\n");
    kill(recv_pid, SIGKILL);
    kill(proc_pid, SIGKILL);
    kill(send_pid, SIGKILL);
}
