/*
 *   (C)opyright 2003 - 2008 RIPE NCC
 *   This file is part of DNSMon
 *
 *   DNSMon is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   DNSMon is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with DNSMon.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "nsmd.h"

/* For poisson */
#define PI 3.141592654

/* For ran1 */
#define IA 16807
#define IM 2147483647
#define AM (1.0/IM)
#define IQ 127773
#define IR 2836
#define NTAB 32
#define NDIV (1+(IM-1)/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)

/* Minimum interval between two measurements, in ms */
#define MIN_INTERVAL 1.00 

float gammaln (float xx);
float ran1 (long *idum); 

extern struct globaltime gt;

/* #define POISSON_DEBUG */

/* 
   Returns as a floating point number an integer value that is a random
   deviate from a Poisson distribution with mean xm, using ran1(idum) as
   a source of random numbers.

   Copied from "Numerical Recipes in C", Section 7.3, page 294.
*/

float
poisson (float xm, long *idum)
{
    static float sq, alxm, g, oldm=(-1.0);
    float em, t, y;
 
    if (xm < 12.0) {
        if (xm != oldm) {
            oldm = xm;
            g = exp (-xm);
        }
        em = -1;
        t = 1.0;
        do {
            ++em;
            t *= ran1(idum);
        } while (t > g);
    } else {
        if ( xm != oldm ) {
            oldm = xm;
            sq = sqrt (2.0 * xm);
            alxm = log(xm);
            g = xm*alxm-gammaln(xm+1);
        }
        do {
            do {
                y = tan (PI * ran1(idum)) ;
                em = sq * y + xm;
            } while (em < 0.0);
                 em = floor (em);
                 t = 0.9 * (1.0+y*y) * exp (em*alxm-gammaln(em+1.0)-g);
        } while (ran1(idum) > t);
    }

    return (em);
}


/* 
   Minimal random number generator of Park and Miller with Bays-Durham shuffle
   and added safe guards.  Returns a uniform random deviate between 0.0 and 1.0
   (exclusive of the end-point values).  Call with idum a negative integer to
   initialize, thereafer do not alter idum in a single sequence.

   Copied from Numerical Recipes in C, page 280, section 7.1
*/

float
ran1 (long *idum) {
    int j;
    long k;
    static long iy=0;
    static long iv[NTAB];
    float temp;

    if (idum <= 0 || !iy) {
        if (-(*idum) < 1) 
            *idum = 1;
        else 
            *idum = -(*idum);
        for (j=NTAB+7; j>= 0; j--) {
            k = (*idum)/IQ;
            *idum = IA*(*idum-k*IQ)-IR*k;
            if (*idum < 0) 
                *idum += IM;
            if (j <NTAB) 
                iv[j] = *idum;
        }
        iy = iv[0];
    }
    k = (*idum) / IQ ;
    *idum = IA * (*idum - k*IQ) - IR*k;
    if (*idum < 0) 
        *idum += IM;
    j = iy/NDIV;
    iy = iv[j];
    iv[j] = *idum;
    if ( (temp=AM*iy) > RNMX) {
        return (RNMX);
    } else 
        return (temp); 
}

/* 
   Returns the value ln(gamma(xx)) for xx>0.

   Copied from Numerical Recipes in C, page 214, section 6.1
*/
float
gammaln (float xx) {
    double x, y, tmp, ser;

    static double cof[6] = {76.18009172947146,
                -86.50532032941677,
                24.01409824083091,
                -1.231739572450155,
                0.1208650973866179e-2,
                -0.5395239384953e-5};
    int j;

    y = x = xx ;
    tmp = x+5.5;
    tmp -= (x+0.5) *log(tmp);
    ser = 1.000000000190015;
    for (j=0 ; j< 5; j++) 
        ser += cof[j]/++y;

    return (-tmp + log(2.5066282746310005 * ser / x));
}

/* 
   Read seed from file: 
      Argument is filename where to find seed, otherwise use default
      Returns seed, or -1 if seed cannot be found.
*/
long poisson_getseed (int debug, char *filename) {

    long seed = 1.0 ; /* Initial seed value */
    FILE *fp;
    char random_seed_file[100] = "./";  /* Where to find the configfile */
    int result;

    if (filename == NULL) {  
        if (getenv ("TTROOT")) {
            strcpy (random_seed_file, getenv("TTROOT"));
        }
        strcat (random_seed_file, "/config/nsmd_poisson.seed");
    } else {
        strcpy (random_seed_file, filename); 
    }

    fp = fopen (random_seed_file, "r");
    if (fp != NULL) {
        result = fscanf (fp, "%ld", &seed);
        if (result != 1) {
            seed = 1;
        }
        fclose (fp);
    } else {
        seed = 1;
    }

    if (debug) {
        printf ("poisson_getseed: read %ld as seed\n", seed);
    }

    return (-1 * seed);
}

/* 
   Write seed to file:
     Argument is filename where to store seed, otherwise use default
     Second argument is seed
   return 0 on success, 1 on failure.
*/
int
poisson_saveseed (int debug, char *filename, long seed) {

    FILE *fp;
    char random_seed_file[100] = "./"; 

    if (filename == NULL) {  
        if (getenv ("TTROOT")) {
            strcpy (random_seed_file, getenv("TTROOT"));
        }
        strcat (random_seed_file, "/config/nsmd_poisson.seed");
    } else {
            strcpy (random_seed_file, filename); 
      }

    if (debug) {
        printf ("poisson_saveseed: writing seed %ld to %s\n",
            seed, random_seed_file);
    }

    fp = fopen (random_seed_file, "w");
    if (fp != NULL) {
        fprintf (fp, "%ld\n", seed);
        fclose (fp);
        return (0);
    } else {
        return (1);
    }
}


/* 
   This generates a sequence of random numbers with intervals according
   to a poisson distribution in target->time.  It does some sanity 
   checking on the sequence.

   Arguments: debug:  TRUE/FALSE, if TRUE be verbose
              target  Struct to file.
              seed:   Random seed, can be overwritten with target->seed
 
   Returns:   seed:   Random seed after the last call.
*/

#define FALSE 0
#define TRUE (!FALSE)

long
do_poisson (int debug, struct server list[], int id, long seed)
{
    long xm;
    long result = 0;
    int linear = FALSE; /* skip poisson */

    /* Mean interval of poisson distribution */
    xm = gt.period / list[id].frequency;
#ifdef POISSON_DEBUG
    printf("mean (xm): %ld\n", xm);
#endif /* POISSON_DEBUG */

    /* Generate interval, make sure that it is above the minimum */
    result = 0;
    while (result < MIN_INTERVAL || result > gt.period) {
        if (linear == FALSE) {
            result = poisson (xm, &seed);
        } else {
            result = xm;
        }
    }
#ifdef POISSON_DEBUG
    printf("result: %ld\n", result);
#endif /* POISSON_DEBUG */

    list[id].time = result;

    return (seed);
}
