/*
 *   (C)opyright 2003 - 2008 RIPE NCC
 *   This file is part of DNSMon
 *
 *   DNSMon is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   DNSMon is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with DNSMon.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <sys/time.h>
#include <sys/types.h>
#include <sys/queue.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/resource.h>
#include <sys/select.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <arpa/nameser.h>
#include <netinet/udp.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#ifdef __FreeBSD__
#include <netinet/ip_var.h>
#endif /* __FreeBSD__ */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>

#ifdef __FreeBSD___
#include <libutil.h>
#endif /* __FreeBSD__ */
#ifdef __linux__
#include "setproctitle.h"
#endif /* __linux__ */


#include "nsmd.h"
#include "dns.h"
#include "receiving_process.h"

int receive_udp_packet(int fd, char *packet);
int receive_raw_packet(char *packet);

int r_pipe[2];
extern int raw_fd;
extern int udp_fd;
#ifdef IPV6_SUPPORT
extern int udp_v6_fd;
#endif /* IPV6_SUPPORT */

/* #define RECEIVING_DEBUG */

pid_t
receiving_process()
{
    char * buf;
    fd_set fdset;
    struct received_packet rpack;
    char * result;
    pid_t pid;
#ifdef IPV6_SUPPORT
    int current_fd;
#endif /* IPV6_SUPPORT */

#ifdef RECEIVING_DEBUG
    printf("%s():\n", __func__);
#endif /* RECEIVING_DEBUG */

    /* setup receiver pipe */
    pipe(r_pipe);

    /* fork of new process */
    if ((pid = fork()))
        return (pid);

    setproctitle("[receiving process]");

#ifdef RECEIVING_DEBUG
    printf("receiver pid: %d\n",getpid());
#endif /* RECEIVING_DEBUG */


    for(;;) {
        FD_ZERO(&fdset);
        FD_SET(udp_fd, &fdset);
        FD_SET(raw_fd, &fdset);
#ifdef IPV6_SUPPORT
        FD_SET(udp_v6_fd, &fdset);
        select(max(max(raw_fd, udp_fd), udp_v6_fd) + 1, &fdset, 
	       NULL, NULL, NULL);

#else
        select(max(raw_fd, udp_fd) + 1, &fdset, NULL, NULL, NULL);
#endif /* IPV6_SUPPORT */

#ifdef IPV6_SUPPORT
        if (FD_ISSET(udp_fd, &fdset) || FD_ISSET(udp_v6_fd, &fdset)) {
            if (FD_ISSET(udp_fd, &fdset)) {
               current_fd = udp_fd;
#ifdef RECEIVING_DEBUG
               printf("Received v4 packet on UDP port\n");
#endif /* RECEIVING_DEBUG */
            } else {
               current_fd = udp_v6_fd;
#ifdef RECEIVING_DEBUG
               printf("Received v6 packet on UDP port\n");
#endif /* RECEIVING_DEBUG */
            }
#else
        if (FD_ISSET(udp_fd, &fdset)) {
#endif /* IPV6_SUPPORT */
            buf = malloc(600);
            result = malloc(100);
            bzero(result, 100);
            bzero(&rpack, sizeof(rpack));
#ifdef IPV6_SUPPORT
            if (receive_udp_packet(current_fd, buf) > 0) {
				/* Set address family in received packet as we want to have different result files for v4/v6/vX */
				rpack.af = current_fd == udp_fd ? AF_INET : current_fd == udp_v6_fd ? AF_INET6 : NULL;
#else
            if (receive_udp_packet(udp_fd, buf) > 0) {
				rpack.af = AF_INET;
#endif /* IPV6_SUPPORT */
                gettimeofday(&rpack.recv_tv, NULL);
                rpack.dns_id = parse_udp_packet(buf, result);
                strcpy(rpack.result, result);
            }
            free(buf);
            free(result);

            if(write(r_pipe[PIPE_WRITE_END], &rpack, 
                sizeof(rpack)) != sizeof(rpack)) {
                printf("ERROR: write to pipe failed\n");
            }

        }

		/* XXX basically dead code as no raw packet processing is performed! */

        if (FD_ISSET(raw_fd, &fdset)) {
            buf = malloc(1600);
            result = malloc(100);
            bzero(&rpack, sizeof(rpack));
            if (receive_raw_packet(buf) > 0) {
				rpack.af = AF_INET; /* XXX Raw sockets only IPv4 at the moment! */
                gettimeofday(&rpack.recv_tv, NULL);
                parse_raw_packet(buf, result);
                rpack.dns_id = -1;
            }
            free(buf);
            free(result);

#ifdef NOTYET
            if(write(r_pipe[PIPE_WRITE_END], &rpack, 
                sizeof(rpack)) != sizeof(rpack)) {
                printf("ERROR: write to pipe failed\n");
            }
#endif /* NOTYET */
        }
    } 

    /* NOT REACHED */
}

/*
 *  Listen on raw socket and record data in buffer
 */
int
receive_raw_packet(char *packet)
{
    int len;
    int ret;
    extern int errno;
    struct sockaddr_in saddr;

    len = sizeof(saddr);
    ret = recvfrom(raw_fd, packet, 1600, 0,
        (struct sockaddr *) &saddr, &len);
    if (ret < 0) {
        printf("ERROR: recvfrom failed!\n");
    }

#ifdef RECEIVING_DEBUG
    printf("%s(): %d bytes received.\n", __func__, ret);
#endif /* RECEIVING_DEBUG */

    /* terminate packet for easier parsing */    
    packet[ret] = 0; 

    return (ret);
}

/*
 *  Listen on udp socket and record data in buffer
 */
int
receive_udp_packet(int fd, char *packet)
{
    int len;
    int ret;
    extern int errno;
    struct sockaddr_in saddr;

    len = sizeof(saddr);
    ret = recvfrom(fd, packet, 512, 0, 
            (struct sockaddr *)&saddr, &len);
    if (ret < 0) {
        printf("ERROR: recvfrom failed!\n");
    }

#ifdef RECEIVING_DEBUG
    printf("%s(): %d bytes received.\n", __func__, ret);
#endif /* RECEIVING_DEBUG */

    /* terminate packet for easier parsing */    
    packet[ret] = 0; 

    return (ret);
}
