/*
 *   (C)opyright 2003 - 2008 RIPE NCC
 *   This file is part of DNSMon
 *
 *   DNSMon is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   DNSMon is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with DNSMon.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <arpa/nameser.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>

/* Use BSD style UDP header on Linux */
#ifdef __linux__
#define __FAVOR_BSD
#endif /* __linux__ */
#include <netinet/udp.h>

#ifdef __FreeBSD__
#include <netinet/ip_var.h>
#endif /* __FreeBSD__ */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <syslog.h>

#ifdef __FreeBSD___
#include <libutil.h>
#endif /* __FreeBSD__ */
#ifdef __linux__
#include "setproctitle.h"
#endif /* __linux__ */


#include "nsmd.h"
#include "scheduler.h"
#include "checksum.h"
#include "dns.h"
#include "sending_process.h"
#include "grimreaper.h"
#include "config.h"


int s_pipe[2];
int g_pipe[2];

extern struct globaltime gt;
extern int raw_fd;

#ifdef IPV6_SUPPORT
extern int udp_v6_fd;
#endif /* IPV6_SUPPORT */

extern unsigned int src_ip;
extern STAILQ_HEAD(, measurement) schedule;
extern int mydebug;
extern int dns_id;

struct timeval sendpacket(char *packet, struct server list[], int entry);

#ifdef IPV6_SUPPORT
struct timeval sendpacket6(char *packet, struct server list[], int entry);
#endif /* IPV6_SUPPORT */

int execute_schedule(struct server list[LIST_SIZE]);
int resend_packet(struct server list[LIST_SIZE],
    struct grimreaper_packet gpack);
int needs_resend(struct server list[LIST_SIZE]);

/* #define SENDING_DEBUG */

pid_t
sending_process()
{
    struct server list[LIST_SIZE];
    pid_t pid;

    pipe(s_pipe);
    pipe(g_pipe);

    /* create separate process */
    if ((pid = fork()))
        return (pid);

    setproctitle("[sending process]");

#ifdef SENDING_DEBUG
    printf("sender pid: %d\n",getpid());
#endif /* SENDING_DEBUG */

    dns_id = getpid();

    for (;;) {
        /* read configuration file */
        read_config(list);

        /* create schedule */
        make_schedule(list);

        /* handle on schedule */
        execute_schedule(list);
    }
}

int
execute_schedule(struct server list[LIST_SIZE])
{
    char *packet;
    struct timeval tv;
    struct measurement *m;
    struct sent_packet spack;
#ifdef IPV6_SUPPORT
    struct in6_addr in6addr;
#endif /* IPV6_SUPPORT */


	if (mydebug) {
		STAILQ_FOREACH(m, &schedule, entries) {
			printf("target: %d scheduled at: %ld type: %d\n", 
				m->targetid, m->time, list[m->targetid].query_type);
		}
	}

    gettimeofday(&tv, NULL);
    gt.start = tv.tv_sec;

    STAILQ_FOREACH(m, &schedule, entries) {
        int sched_id = m->targetid;
        long sleeptime = 0;

        /* see if there are packets to resend */
        needs_resend(list);

        gettimeofday(&tv, NULL);
        if ((m->time + gt.start) < tv.tv_sec) {
            if (mydebug) {
                printf("ERROR: missed a measurement,"
                    "go fix your scheduler!!\n");
                printf("m->time: %ld gt.start: %ld "
                    "tv.tv_sec: %ld\n", m->time, 
                    gt.start, tv.tv_sec);
            } else {
                syslog(LOG_DEBUG, 
                    "ERROR: missed a measurement, "
                    "go fix your scheduler!!\n");
                syslog(LOG_DEBUG, 
                    "m->time: %ld gt.start: %ld "
                    "tv.tv_sec: %ld\n",
                    m->time, gt.start, tv.tv_sec);
            }
        }

        if ( (m->time + gt.start) > tv.tv_sec) {
            /* we are scheduled LATER */
			sleeptime = (((m->time + gt.start) - tv.tv_sec) * 1000000 ) - tv.tv_usec;

            if (mydebug)
				printf("going to sleep for: %ld ms\n", sleeptime/1000);
            usleep(sleeptime);
        } 

        packet = (char *)malloc(512);
        bzero(&spack, sizeof(spack));
        spack.dns_id = buildpacket(packet, list, sched_id);

#ifdef IPV6_SUPPORT
        if(inet_pton(AF_INET6, list[sched_id].address, &in6addr)) {
			if(candov6) {
				spack.sent_tv = sendpacket6 (packet, list, sched_id);
				spack.af = AF_INET6;
#ifdef SENDING_DEBUG
				printf ("Address %s is considered IPv6\n", list[sched_id].address);
			} else {
				printf ("Address %s is considered IPv6, but host has no IPv6 configured\n", list[sched_id].address);
#endif /* SENDING_DEBUG */
			}
        } else {
            /* this is probably a IPv4 address */
            spack.sent_tv = sendpacket (packet, list, sched_id);
			spack.af = AF_INET;
#ifdef SENDING_DEBUG
            printf ("Address %s is considered IPv4\n", list[sched_id].address);
#endif /* SENDING_DEBUG */
        }
#else
        spack.sent_tv = sendpacket(packet, list, sched_id);
		spack.af = AF_INET;
#endif /* IPV6_SUPPORT */

#ifdef SENDING_DEBUG
        printf("Sending with sched_id: %d\n", sched_id);
#endif /* SENDING_DEBUG */

		if (spack.af) { /* AF_INET and AF_INET6 are both non-zero */

			/* 
			 * Only log packet when spack.af is set. this can be false when we
			 * have a IPv6 target on a host with a IPv6 stack, but no IPv6 configured 
			 */

			spack.sched_id = sched_id;
			spack.query_type = list[sched_id].query_type;
			spack.timeout = list[sched_id].timeout;
			spack.attempts = 1;
			strcpy(spack.address, list[sched_id].address);
			strcpy(spack.label, list[sched_id].label);
			strcpy(spack.domain, list[sched_id].domain);
			strcpy(spack.qry, list[sched_id].qry);
			spack.server_ip = list[sched_id].server_ip;

			if(write(s_pipe[PIPE_WRITE_END], &spack, sizeof(spack)) != 
				sizeof(spack)) {
				printf("ERROR: write to pipe failed\n");
			}
		}

        free(packet); 

    } /* stailq_foreach measurement */

    gettimeofday(&tv, NULL);
    if (tv.tv_sec < (gt.start + gt.period)) {
        /* sleep until the end of the period */
#ifdef SENDING_DEBUG
        printf("we have to wait until the period has finished\n");
        printf("now: %ld start+period: %ld\n", tv.tv_sec, 
            gt.start + gt.period);
#endif /* SENDING_DEBUG */
        sleep((gt.start + gt.period) - tv.tv_sec);
    }

    return (0);
}

/*
 *  Sendpacket
 *  takes a buffer and send it out to socket
 */
struct timeval
sendpacket(char *packet, struct server list[], int entry)
{
    int ret;
    int size;
    struct ip *iphdr;
    struct udphdr *udphdr;
    char *sndbuf = (char *)malloc(512);
    int udplen;
    int iplen;
    int datalen;
    struct sockaddr_in saddr;
    struct timeval send_time;

    datalen = list[entry].size;

    iplen = sizeof(struct ip);
    udplen = sizeof(struct udphdr);

    iphdr = (struct ip *) sndbuf;
    udphdr = (struct udphdr *) (sndbuf + sizeof(struct ip));

    /* copy to the actual send buffer */
    memcpy(sndbuf+iplen+udplen, packet, datalen);

    iphdr->ip_v = IPVERSION;
        iphdr->ip_hl=sizeof(struct ip) >> 2;
    iphdr->ip_id = 0;  /* 0 means kernel set appropriate value */
    iphdr->ip_len = sizeof(struct ip) + sizeof(struct udphdr) + datalen;
    iphdr->ip_tos = 0;
        iphdr->ip_off = 0;
        iphdr->ip_p = IPPROTO_UDP;
    iphdr->ip_ttl = 0x40;

    /* set aquired first ip address */
        iphdr->ip_src.s_addr = htonl(src_ip);
        iphdr->ip_dst.s_addr = inet_addr(list[entry].address);

    size = iplen + udplen + datalen;

    udphdr = (struct udphdr *) (sndbuf + sizeof(struct ip));

    udphdr->uh_sport = htons(UDP_SRC_PORT);
    udphdr->uh_dport = htons(53);
    udphdr->uh_ulen = htons(sizeof(struct udphdr) + datalen);
    udphdr->uh_sum = do_checksum(sndbuf, IPPROTO_UDP, 
        sizeof(struct udphdr) + datalen);


    /* this will be the time the measurements get logged */
    gettimeofday(&send_time, NULL);

    ret = sendto(raw_fd, sndbuf, size, 0, 
        (struct sockaddr *)&saddr, sizeof(saddr));
    if (ret < 0) {
        perror("sendto");
        exit(1);
    }
#ifdef SENDING_DEBUG
    printf("%d bytes send.\n", ret);
#endif /* SENDING_DEBUG */

    free(sndbuf);

    return (send_time);
}

#ifdef IPV6_SUPPORT
/*
 *  Sendpacket6
 *  takes a buffer and send it out to IPv6 UDP socket
 */
struct timeval
sendpacket6(char *packet, struct server list[], int entry)
{
    int ret;
    int datalen;
    struct sockaddr_in6 saddr;
    struct timeval send_time;

    datalen = list[entry].size;

    saddr.sin6_len = sizeof(struct sockaddr_in6);
    saddr.sin6_family = AF_INET6;
    saddr.sin6_port = htons(53);
    saddr.sin6_flowinfo = 0;
    inet_pton(AF_INET6, list[entry].address, &saddr.sin6_addr);

    /* this will be the time the measurements get logged */
    gettimeofday(&send_time, NULL);

    ret = sendto(udp_v6_fd, packet, datalen, 0, 
        (struct sockaddr *)&saddr, sizeof(saddr));
    if (ret < 0) {
        perror("sendto");
        exit(1);
    }
#ifdef SENDING_DEBUG
    printf("%d bytes send.\n", ret);
#endif /* SENDING_DEBUG */

    return (send_time);
}
#endif /* IPV6_SUPPORT */

int
needs_resend(struct server list[LIST_SIZE])
{
    fd_set fdset;
    struct timeval timeout;
    struct grimreaper_packet gpack;

    FD_ZERO(&fdset);
    FD_SET(g_pipe[PIPE_READ_END], &fdset);
    timeout.tv_sec = 0; timeout.tv_usec = 0;
    select(g_pipe[PIPE_READ_END] + 1, &fdset, NULL, NULL, &timeout);

    if (!FD_ISSET(g_pipe[PIPE_READ_END], &fdset))
        return (0);

    bzero(&gpack, sizeof(gpack));
    if (read(g_pipe[PIPE_READ_END], &gpack, sizeof(gpack))
        != sizeof(gpack)) {
        printf("ERROR: read to pipe failed\n");
    }

    resend_packet(list, gpack);

    return (0);
}

int
resend_packet(struct server list[LIST_SIZE], struct grimreaper_packet gpack)
{
    char * packet = (char *) malloc(512);
    struct sent_packet spack;
    int sched_id;

#ifdef IPV6_SUPPORT
    struct in6_addr in6addr;
#endif /* IPV6_SUPPORT */

    sched_id = gpack.sched_id;

    bzero(&spack, sizeof(spack));
    spack.dns_id = buildpacket(packet, list, sched_id);

#ifdef IPV6_SUPPORT
        if(inet_pton(AF_INET6, list[sched_id].address, &in6addr)) {
			if(candov6) {
				spack.sent_tv = sendpacket6 (packet, list, sched_id);
				spack.af = AF_INET6;
#ifdef SENDING_DEBUG
				printf ("Address %s is considered IPv6\n", list[sched_id].address);
			} else {
				printf ("Address %s is considered IPv6, but host has no IPv6 configured\n", list[sched_id].address);
#endif /* SENDING_DEBUG */
			}
        } else {
            /* this is probably a IPv4 address */
            spack.sent_tv = sendpacket (packet, list, sched_id);
			spack.af = AF_INET;
#ifdef SENDING_DEBUG
            printf ("Address %s is considered IPv4\n", list[sched_id].address);
#endif /* SENDING_DEBUG */
        }
#else
        spack.sent_tv = sendpacket(packet, list, sched_id);
		spack.af = AF_INET;
#endif /* IPV6_SUPPORT */

#ifdef SENDING_DEBUG
        printf("Resending with sched_id: %d\n", sched_id);
#endif /* SENDING_DEBUG */

	if (spack.af) { /* AF_INET and AF_INET6 are both non-zero */

		/* 
		 * Only log packet when spack.af is set. this can be false when we
		 * have a IPv6 target on a host with a IPv6 stack, but no IPv6 configured 
		 */
		spack.sched_id = sched_id;
		spack.query_type = list[sched_id].query_type;
		spack.timeout = list[sched_id].timeout;
		spack.attempts = ++(gpack.attempts);
		strcpy(spack.address, list[sched_id].address);
		strcpy(spack.label, list[sched_id].label);
		strcpy(spack.domain, list[sched_id].domain);
		strcpy(spack.qry, list[sched_id].qry);
		spack.server_ip = list[sched_id].server_ip;

		if(write(s_pipe[PIPE_WRITE_END], &spack, sizeof(spack)) != 
			sizeof(spack)) {
			printf("ERROR: write to pipe failed\n");
		}
	}

    free(packet); 

    return (0);
}
